import java.util.*;
import java.io.*;

class InsolublePuzzleError extends Exception {
}

class InvalidPuzzleError extends Exception {
}

/* This class contains all the data about a puzzle */
public class Puzzle {
	int[] PuzzleRows;
	int[] SolutionRows;
	int[] AttemptedRows;
	String SourceDate = "Unknown";
	String CreationDate = "Unknown";
	String Author = "Unknown";
	String Difficulty = "Unknown";
	String Source = "Unknown";
	int GridType;
	String[][] Notes;
	
	static int FourXTwo = 4;	
	static int ThreeXThree = 0;
	static int FourXFour = 1;
	static int FiveXFive = 2;
	static int TwoXTwo = 3;
	
	static String[] Type = {"3x3", "4x4", "5x5", "2x2", "4x2"};
	
	public Puzzle() {
		this(2);
	}
	
	public Puzzle(int PuzzleType) {
		GridType = PuzzleType;
		updateType();
	}
	
	public void setBlank() {
		for (int i = 0; i < PuzzleArrays.Sizes[GridType]; i++) {
			PuzzleRows[i] = 0;
			SolutionRows[i] = 0;
			AttemptedRows[i] = 0;
		}
	}
	
	public String getType() {
		return Type[GridType];
	}
	
	public int getGridType() {
		return GridType;
	}
	
	public String getDifficulty() {
		return Difficulty;
	}
	
	public String getSource() {
		return Source;
	}
	
	public String getAuthor() {
		return Author;
	}
	
	public String getSourceDate() {
		return SourceDate;
	}
	
	public void setSourceDate(String s) {
		SourceDate = s;
	}
	
	public void setSource(String s) {
		Source = s;
	}
	
	public void setDifficulty(String s) {
		Difficulty = s;
	}
	
	private void updateType() {		
		PuzzleRows = new int[PuzzleArrays.Sizes[GridType]];
		SolutionRows = new int[PuzzleArrays.Sizes[GridType]];
		AttemptedRows = new int[PuzzleArrays.Sizes[GridType]];
		Notes = new String[PuzzleArrays.RowLength[GridType] + 1][];
		for (int i = 0; i < PuzzleArrays.Sizes[GridType]; i++) {
			PuzzleRows[i] = 0;
			SolutionRows[i] = 0;
			AttemptedRows[i] = 0;
		}
		/* The + 1 means I can use e.g. 1-9 rather than 0-8 which is easier conceptually */
		for (int i = 0; i < PuzzleArrays.RowLength[GridType] + 1; i++) {
			Notes[i] = new String[PuzzleArrays.ColumnLength[GridType] + 1];
		}
	}
	
	public void setType(int T) {
		if (GridType != T) {
			GridType = T;
			updateType();
		}
	}
	
	public void setType(String S) {
		int Match = -1; // -1 is impossible so useful for not found
		for (int i = 0; i < Type.length; i++) {
			if (S.equals(Type[i])) {
				Match = i;				
			}
		}
		if (Match != -1) {
			if (GridType != Match) {
				GridType = Match;
				updateType();
			}
		}
	}
	
	public void setAuthor(String s) {
		Author = s;
	}
	
	public String getCreationDate() {
		return CreationDate;
	}
	
	public void setCreationDate(String s) {
		CreationDate = s;
	}
	
	/* This determines if the puzzle is of the right size by checking that they all match one of the allowable lengths */
	private void checkPuzzleLength() throws InvalidPuzzleError {
		boolean Match = false;
		for (int i = 0; i < PuzzleArrays.Sizes.length; i++) {
			if (PuzzleRows.length == PuzzleArrays.Sizes[i]) {
				GridType = i;
				Match = true;
				if (SolutionRows.length != PuzzleArrays.Sizes[GridType]) {
					SolutionRows = new int[PuzzleArrays.Sizes[GridType]];
					for (int j = 0; j < PuzzleArrays.Sizes[GridType]; j++) {
						SolutionRows[j] = 0;
					}
				}
				if (AttemptedRows.length != PuzzleArrays.Sizes[GridType]) {
					AttemptedRows = new int[PuzzleArrays.Sizes[GridType]];
					for (int j = 0; j < PuzzleArrays.Sizes[GridType]; j++) {
						AttemptedRows[j] = 0;
					}
				}
			}
		}
		if (!Match) throw new InvalidPuzzleError();
	}
	
	public void setPuzzle(int[] P) throws InvalidPuzzleError {
		PuzzleRows = P;
		checkPuzzleLength();
	}
	
	/* Used when reading in puzzles to parse off the details */
	public void setPuzzle(String s) throws InvalidPuzzleError {
		String number = "";
		Vector Entries = new Vector();
	
		for (int i = 0; i < s.length(); i++) {
			if (!(s.charAt(i) == ' ')) {
				number += s.charAt(i);
			} else {
				if (!(number.equals(""))) {
					Entries.add(new Integer(number));
					number = "";
				}
			}
		}
		
		if (!(number.equals(""))) {
			Entries.add(new Integer(number));
		}
		
		PuzzleRows = new int[Entries.size()];
		for (int i = 0; i < Entries.size(); i++) {
			PuzzleRows[i] = ((Integer) Entries.get(i)).intValue();
		}
		checkPuzzleLength();
	}
	
	public int[] getPuzzle() {
		return PuzzleRows;
	}
	
	/* Writes out the current puzzle. Used mainly for testing */
	public void printPuzzle() {
		if (PuzzleRows.length == PuzzleArrays.Sizes[GridType]) {
			for (int i = 0; i < PuzzleArrays.Sizes[GridType]; i++) {
				System.out.print("\t" + PuzzleRows[i]);
				if (i % PuzzleArrays.RowLength[GridType] == (PuzzleArrays.RowLength[GridType] - 1)) {
					System.out.println();
				}
			}		
		}
	}
	
	public boolean solvePuzzle() throws InsolublePuzzleError {
		try {
			// This actually does the hard work of solving
			Worker W = new Worker(getPuzzle());
			if (W.solve()) {
				setSolution(W.getSolution());
				return true;
			} else {
				return false;
			}
		}
		catch (Exception e) {
			e.printStackTrace();
			throw new InsolublePuzzleError();
		}
	}
	
	public void saveToFile(PrintWriter pw) {
		pw.println("<Sudoku>");
		pw.println("<Puzzle>");
		for (int i = 0; i < PuzzleRows.length; i++) {
			pw.print(PuzzleRows[i] + " ");
		}
		pw.println();
		pw.println("</Puzzle>");
		pw.println("<Solution>");
		for (int i = 0; i < SolutionRows.length; i++) {
			pw.print(SolutionRows[i] + " ");
		}
		pw.println();
		pw.println("</Solution>");
		pw.println("<Attempted>");
		for (int i = 0; i < AttemptedRows.length; i++) {
			pw.print(AttemptedRows[i] + " ");
		}
		pw.println();
		pw.println("</Attempted>");
		if (!SourceDate.equals("")) {
			pw.println("<SourceDate>");
			pw.println(SourceDate);
			pw.println("</SourceDate>");
		}
		pw.println("<Type>");
		pw.println(Type[GridType]);
		pw.println("</Type>");
		if (!Author.equals("")) {
			pw.println("<Author>");
			pw.println(Author);
			pw.println("</Author>");
		}
		if (!Source.equals("")) {
			pw.println("<Source>");
			pw.println(Source);
			pw.println("</Source>");
		}
		if (!CreationDate.equals("")) {
			pw.println("<CreationDate>");
			pw.println(CreationDate);
			pw.println("</CreationDate>");
		}
		if (!Difficulty.equals("")) {
			pw.println("<Difficulty>");
			pw.println(Difficulty);
			pw.println("</Difficulty>");
		}
		pw.println("<Notes>");
		for (int i = 0; i < Notes.length; i++) {
			for (int j = 0; j < Notes[i].length; j++) {
				if ((Notes[i][j] != "") && (Notes[i][j] != null)) {
					pw.println("<Note X=\"" + i + "\" Y=\"" + j + "\">" + Notes[i][j] + "</Note>");
				}
			}
		}
		pw.println("</Notes>");
		pw.println("</Sudoku>");
	} 
	
	public void setSolution(int[] S) {
		SolutionRows = S;
	}
	
	/* Used when reading from file */ 
	public void setSolution(String s) {
		String number = "";
		Vector Entries = new Vector();
	
		for (int i = 0; i < s.length(); i++) {
			if (!(s.charAt(i) == ' ')) {
				number += s.charAt(i);
			} else {
				if (!(number.equals(""))) {
					Entries.add(new Integer(number));
					number = "";
				}
			}
		}
		
		if (!(number.equals(""))) {
			Entries.add(new Integer(number));
		}
		
		SolutionRows = new int[Entries.size()];
		for (int i = 0; i < Entries.size(); i++) {
			SolutionRows[i] = ((Integer) Entries.get(i)).intValue();
		}
	}
	
	/* Used when reading from file */
	public void setAttempted(String s) {
		String number = "";
		Vector Entries = new Vector();
	
		for (int i = 0; i < s.length(); i++) {
			if (!(s.charAt(i) == ' ')) {
				number += s.charAt(i);
			} else {
				if (!(number.equals(""))) {
					Entries.add(new Integer(number));
					number = "";
				}
			}
		}
		
		if (!(number.equals(""))) {
			Entries.add(new Integer(number));
		}
		
		AttemptedRows = new int[Entries.size()];
		for (int i = 0; i < Entries.size(); i++) {
			AttemptedRows[i] = ((Integer) Entries.get(i)).intValue();
		}
	}
	
	public void setAttempted(int[] Attempted) {
		AttemptedRows = Attempted;
	}
	
	public int[] getAttempted() {
		return AttemptedRows;
	}
	
	public int[] getSolution() {
		return SolutionRows;
	}
	
	public boolean AttemptedExists() {
		for (int i = 0; i < PuzzleArrays.Sizes[GridType]; i++) {
			if ((AttemptedRows[i] != PuzzleRows[i]) && (AttemptedRows[i] != 0)) {
				return true;
			}
		}
		return false;
	}
	
	public void setNotes(int X, int Y, String Note) {
		Notes[X][Y] = Note;
	}
	
	public String getNotes(int X, int Y) {
		return Notes[X][Y];
	}
}